package gatewayapi.sample.iaszone;

import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.function.BiConsumer;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import com.mmbnetworks.gatewayapi.GatewayClient;
import com.mmbnetworks.gatewayapi.entity.Device;
import com.mmbnetworks.gatewayapi.entity.Property;
import com.mmbnetworks.rapidconnectdevice.zcl.ZigBee;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.JSONUtils;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.MessageType;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.PropertyNames;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.ZCLUnicastMessage;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.ZCLUnicastMessage.ZCLUnicastMessageBuilder;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.ZigBeeMessageTypeAdapter;

public class IasProtocolPassthroughHandler implements BiConsumer<Device, String> {
    
    private static final String IAS_ZONE_CLUSTER_ID = "0500";
    private static final String ZONE_ENROLL_REQUEST_ID = "01";
    private static final String ZONE_ENROLL_RESPONSE_ID = "00";
    
    private final JsonParser jsonParser;
    
    public IasProtocolPassthroughHandler() {
        this.jsonParser = new JsonParser();
    }

    @Override
    public void accept(Device device, String json) {
        // This method will be called whenever an underlying protocol message is being passed through.
        // By default, client callbacks are called from a single thread managed by GAPI.
        // Therefore, avoid writing blocking code in this method, or other client callbacks
        // (such as DeviceEventHandlers, or PropertyUpdateHandlers) may be blocked.

        // See the ProtocolPassthrough documentation for the format of the json message,
        // or inspect the ZigBeeMessageTypeAdapter class for possible key names.
        try {
            final JsonObject jsonObject = jsonParser.parse(json).getAsJsonObject();
            final String messageType = jsonObject.get(ZigBeeMessageTypeAdapter.MESSAGE_TYPE_KEY).getAsString();
            final JsonObject message = jsonObject.get(ZigBeeMessageTypeAdapter.MESSAGE_KEY).getAsJsonObject();

            handleClusterCommands(device, messageType, message);

        } catch (Exception ex) {
            System.out.println("Exception handling protocol msg: " + json + ", Exception: " + ex.toString());
        }
    }
    
    private void handleClusterCommands(final Device device, final String messageType, final JsonObject message) {
        if (  messageType.equalsIgnoreCase(MessageType.ZCL_RESPONSE.toString())
           || messageType.equalsIgnoreCase(MessageType.ZCL_PASSTHROUGH.toString())) {
            
            final boolean isClusterCommand = JSONUtils.parseBitmap8(message.get(ZigBeeMessageTypeAdapter.FRAME_CONTROL_KEY).getAsString())
                    .isBitSet(ZigBee.FrameConstants.FRAME_CONTROL_BIT_POS_CLUSTER_COMMAND.toInt());

            if (isClusterCommand) {
                handleZoneEnrollRequest(device, message);
            }   
        }
    }

    private void handleZoneEnrollRequest(final Device device, final JsonObject message) {
        final String clusterId = message.get(ZigBeeMessageTypeAdapter.CLUSTER_ID_KEY).getAsString(); 
        if (clusterId.equalsIgnoreCase(IAS_ZONE_CLUSTER_ID)) {
            final String commandId = message.get(ZigBeeMessageTypeAdapter.COMMAND_ID_KEY).getAsString();
            if (commandId.equalsIgnoreCase(ZONE_ENROLL_REQUEST_ID)) {
                System.out.println("Handling Zone Enroll Request");
                zoneEnrollResponse(device).thenAccept( jsonResponse -> {
                    System.out.println(jsonResponse);
                });
            }
        }
    }
    
    private CompletableFuture<String> zoneEnrollResponse(final Device device) {
        // This is an example of using the Protocol Passthrough API, with a zigbee message formed as a json string.
        // The zigbee message we're forming is a Zone Enroll Response Command (see ZCL Spec Section 8.2.2.3.1)
        try {
            final Optional<Property> protocolProperty = device.getCachedProperty(PropertyNames.PROTOCOL_PROPERTY_NAME);
            if (protocolProperty.isPresent()) {
                final Property p = protocolProperty.get();
                final JsonObject jsonObject = jsonParser.parse(p.getValue()).getAsJsonObject();
                
                ZCLUnicastMessageBuilder builder = new ZCLUnicastMessageBuilder();
                builder.setGatewayAPIVersion(GatewayClient.getApiVersion());
                builder.setProtocolName("zigbee");
                builder.setProtocolVersion(3);
                builder.setNodeID(jsonObject.get(ZigBeeMessageTypeAdapter.NODE_ID_KEY).getAsString());
                builder.setEndpointID(jsonObject.get(ZigBeeMessageTypeAdapter.ENDPOINT_ID_KEY).getAsString());
                builder.setClusterID(IAS_ZONE_CLUSTER_ID);
                builder.setResponseOptions(ZigBee.FrameConstants.RESPONSE_OPTIONS_APS_RESPONSE.toBitmap8());
                builder.setFrameControl(ZigBee.FrameConstants.FRAME_CONTROL_CLIENT_TO_SERVER_CLUSTER_CMD.toBitmap8());
                builder.setCommandID(ZONE_ENROLL_RESPONSE_ID);
                // Zone Enroll Response:
                // Field 1: 00 - enroll success
                // Field 2: 01 - zone id
                // therefore, final payload is: 0x0001 
                builder.setPayload("0x0001");
                ZCLUnicastMessage message = builder.build();
                
                System.out.println("Sending Zone Enroll Response");
                return device.sendProtocolMessage(message.toJson());    
            }
            
        } catch (Exception ex) {
            System.out.println("error enrolling device: " + ex.getMessage());
        }
        
        CompletableFuture<String> c = new CompletableFuture<>();
        c.completeExceptionally(new Throwable("error enrolling device"));
        return c;
    }

}
